USE master
GO

-- Create three server logins
CREATE LOGIN Rob WITH PASSWORD = 'jackpot_0'
CREATE LOGIN Tammie WITH PASSWORD = 'jackpot_0'
CREATE LOGIN Vince WITH PASSWORD = 'jackpot_0'
GO

-- Create a new database
IF EXISTS(SELECT name FROM sys.databases WHERE name = 'MyDB')
 DROP DATABASE MyDB
GO
CREATE DATABASE MyDB
GO
USE MyDB
GO

-- Create three database users mapped to the logins
CREATE USER Rob FOR LOGIN Rob
CREATE USER Tammie FOR LOGIN Tammie
CREATE USER Vince FOR LOGIN Vince
GO

-- Create a schema owned by Rob
CREATE SCHEMA Sales AUTHORIZATION Rob 
GO

-- Create a table in the schema owned by Rob
CREATE TABLE Sales.Leads (id int, name varchar(50), phone varchar(20)) 

-- Allow Tammie but not Vince to SELECT from the table in the schema owned by Rob
GRANT SELECT ON Sales.Leads TO Tammie 
DENY SELECT ON Sales.Leads TO Vince 
GO

-- Transfer ownership of Sales schema to Tammie
ALTER AUTHORIZATION ON SCHEMA::Sales TO Tammie

-- Cleanup
DROP TABLE Sales.Leads
DROP SCHEMA Sales
DROP USER Vince
DROP USER Tammie
DROP USER Rob
USE master
GO
DROP LOGIN Rob
DROP LOGIN Tammie
DROP LOGIN Vince
GO

-- Execution context example
USE master
GO

-- Create 3 logins
CREATE LOGIN Login1 WITH PASSWORD = 'P@$$w0rd1'
CREATE LOGIN Login2 WITH PASSWORD = 'P@$$w0rd2'
CREATE LOGIN Login3 WITH PASSWORD = 'P@$$w0rd3'
GO 

-- Create a new database
IF EXISTS(SELECT name FROM sys.databases WHERE name = 'MyDB')
 DROP DATABASE MyDB
GO
CREATE DATABASE MyDB
GO
USE MyDB
GO

-- Create 3 users mapped to the 3 logins
CREATE USER User3 FOR LOGIN Login3
CREATE USER User2 FOR LOGIN Login2
CREATE USER User1 FOR LOGIN Login1
GO 

-- Create a corresponding schema for each of the 3 users
CREATE SCHEMA Schema3 AUTHORIZATION User3 
GO 
CREATE SCHEMA Schema2 AUTHORIZATION User2
GO 
CREATE SCHEMA Schema1 AUTHORIZATION User1
GO 

-- Let User3 create tables and let User2 create stored procedures
GRANT CREATE TABLE TO User3 
GRANT CREATE PROCEDURE TO User2 
GO 

-- Impersonate Login3 (User3)
EXECUTE AS LOGIN = 'Login3' 
GO 

-- Create and populate a table in Schema3
CREATE TABLE Schema3.Region(RegionName nvarchar(50)) 
INSERT INTO Schema3.Region VALUES('East Coast'), ('West Coast') , ('Midwest') 

-- Allow User2 to SELECT from the Schema3 table
GRANT SELECT ON Schema3.Region TO User2 
GO 
REVERT 
GO

-- Impersonate Login2 (User2)
EXECUTE AS LOGIN = 'Login2' 
GO

-- Create a stored procedure in Schema2 that selects from the table in Schema3
CREATE PROCEDURE Schema2.GetRegions AS 
 SELECT * FROM Schema3.Region 
GO 

-- Allow User1 to EXECUTE the Schema2 stored procedure
GRANT EXECUTE ON Schema2.GetRegions TO User1 
GO 
REVERT
GO

-- User1 cannot access the Schema3 table, even via a Schema2 stored
-- proc that they have permission to execute
EXECUTE AS LOGIN = 'Login1' 
GO 
SELECT * FROM Schema3.Region  -- fails
EXEC Schema2.GetRegions       -- fails
GO 
REVERT 
GO
 
-- Modify the Schema2 stored proc so that it always run in the context
-- of the owner, not the caller
EXECUTE AS LOGIN = 'Login2'
GO 
ALTER PROCEDURE Schema2.GetRegions WITH EXECUTE AS OWNER AS 
 SELECT * FROM Schema3.Region 
GO 
REVERT 
GO

-- User1 still cannot access the Schema3 table directly, but now they can
-- indirectly via the Schema2 stored procedure
EXECUTE AS LOGIN = 'Login1' 
GO
SELECT * FROM Schema3.Region  -- fails
EXEC Schema2.GetRegions       -- works!
GO 
REVERT 
GO

-- Cleanup
USE master
GO

DROP LOGIN Login1
DROP LOGIN Login2
DROP LOGIN Login3

DROP DATABASE MyDB
