﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Data.EntityClient;

namespace DemoEntityFramework
{
  public partial class Form1 : Form
  {
    public Form1()
    {
      InitializeComponent();
    }

    private void btnLinqWithDirectSql_Click(object sender, EventArgs e)
    {
      // Retrieve via direct SQL:
      using (var ctx = new SampleDbEntities())
      {
        // Define, but do not execute, the query (deferred execution)
        //  'q' is of type System.Data.Objects.ObjectQuery<Customer>
        var q =
          from customer in ctx.Customers
          where !customer.LastName.StartsWith("A")
          orderby customer.LastName descending
          select customer;

        // Implicitly execute query on the server, returning List<Customer>
        var list = q.ToList();

        MessageBox.Show(string.Format("Non-A customer count: {0}", list.Count));
      }
    }

    private void btnLinqWithStoredProc_Click(object sender, EventArgs e)
    {
      // Retrieve via a function import mapped to a stored procedure:
      using (var ctx = new SampleDbEntities())
      {
        // Execute the stored procedure on the server immediately
        //  'q' is of type System.Linq.OrderedEnumerable<Customer>
        var q =
          from customer in ctx.SelectCustomers()
          where !customer.LastName.StartsWith("A")
          orderby customer.LastName descending
          select customer;

        // Filter and sort stored proc results on the client, returning List<Customer>
        var list = q.ToList();

        MessageBox.Show(string.Format("Non-A customer count: {0}", list.Count));
      }
    }

    private void btnUpdatingData_Click(object sender, EventArgs e)
    {
      using (var ctx = new SampleDbEntities())
      {
        var customers =
          (from cust in ctx.Customers
           where !cust.LastName.StartsWith("A")
           orderby cust.LastName descending
           select cust).ToList();

        if (customers.Count == 0)
        {
          MessageBox.Show("There are no customers");
          return;
        }

        var firstCustomer = customers[0];
        var orderCount = firstCustomer.OrderHeaders.Count;
        MessageBox.Show(string.Format(
          "Customer {0} has {1} order(s)", firstCustomer.CustomerId, orderCount));

        // Change the customer's name
        firstCustomer.FirstName = "Keith";
        firstCustomer.LastName = "Harris";

        // Add a customer order
        firstCustomer.OrderHeaders.Add(new OrderHeader()
        {
          ShipVia = "Regular Mail",
          OrderStatus = "Open",
        });

        ctx.SaveChanges();
      }
    }

    private void btnManyToMany_Click(object sender, EventArgs e)
    {
      using (var ctx = new SampleDbEntities())
      {
        // Create 3 customers and 3 employees

        var custClaus = new Customer() { FirstName = "Claus", LastName = "Hansen" };
        var custTerry = new Customer() { FirstName = "Terry", LastName = "Adams" };
        var custDan = new Customer() { FirstName = "Dan", LastName = "Park" };

        ctx.Customers.AddObject(custClaus);
        ctx.Customers.AddObject(custTerry);
        ctx.Customers.AddObject(custDan);

        var empDavid = new Employee() { FirstName = "David", LastName = "Alexander" };
        var empAndy = new Employee() { FirstName = "Andy", LastName = "Jacobs" };
        var empErik = new Employee() { FirstName = "Erik", LastName = "Ryan" };

        ctx.Employees.AddObject(empDavid);
        ctx.Employees.AddObject(empAndy);
        ctx.Employees.AddObject(empErik);

        // Create these many-to-many relationships:
        //  Customer : Employees         | Employee : Customers
        //   Claus   :  Andy, Erik       |  Andy    :  Claus, Terry
        //   Terry   :  Andy, David      |  David   :  Terry, Dan
        //   Dan     :  David, Erik      |  Erik    :  Claus, Dan

        // Can either add employees to customers... 

        custClaus.Employees.Add(empAndy);
        custClaus.Employees.Add(empErik);

        custTerry.Employees.Add(empAndy);
        custTerry.Employees.Add(empDavid);

        // ...or add customers to employees

        empErik.Customers.Add(custDan);
        empDavid.Customers.Add(custDan);

        // Save new customers, employees, and their relationships to the database
        ctx.SaveChanges();

        // Get customers belonging to employee David (Terry and Dan)
        this.lstCustomers.Items.Clear();
        foreach (var cust in empDavid.Customers)
        {
          this.lstCustomers.Items.Add(
            string.Format("{0} {1}", cust.FirstName, cust.LastName));
        }

        // Get employees belonging to customer Terry (Andy and David)
        this.lstEmployees.Items.Clear();
        foreach (var emp in custTerry.Employees)
        {
          this.lstEmployees.Items.Add(
            string.Format("{0} {1}", emp.FirstName, emp.LastName));
        }
      }
    }

    private void btnEntitySql_Click(object sender, EventArgs e)
    {
      const string EntitySql = @"
        SELECT VALUE cust
          FROM SampleDbEntities.Customers AS cust
          WHERE cust.LastName NOT LIKE 'A%'
          ORDER BY cust.LastName";

      using (var ctx = new SampleDbEntities())
      {
        // Define, but do not execute, the Entity SQL query (deferred execution)
        var q = ctx.CreateQuery<Customer>(EntitySql);

        // Implicitly execute query on the server, returning List<Customer>
        var list = q.ToList();

        MessageBox.Show(string.Format("Non-A customer count: {0}", list.Count));
      }
    }

    private void btnEntityClient_Click(object sender, EventArgs e)
    {
      const string ConnStr = "name=SampleDbEntities";

      const string EntitySql = @"
        SELECT VALUE cust
          FROM SampleDbEntities.Customers AS cust
          WHERE cust.LastName NOT LIKE @Pattern
          ORDER BY cust.LastName";

      var names = new List<string>();
      using (var conn = new EntityConnection())
      {
        conn.ConnectionString = ConnStr;
        conn.Open();
        using (var cmd = new EntityCommand())
        {
          cmd.Connection = conn;
          cmd.CommandText = EntitySql;
          cmd.Parameters.AddWithValue("Pattern", "A%");
          using (var rdr = cmd.ExecuteReader(CommandBehavior.CloseConnection |
            CommandBehavior.SequentialAccess))
          {
            while (rdr.Read())
            {
              var firstName = rdr.GetString(1);
              var lastName = rdr.GetString(2);
              var name = string.Format("{0}, {1}", lastName, firstName);
              names.Add(name);
            }
            rdr.Close();
          }
        }
        this.lstCustomers.DataSource = names;
      }
    }

  }
}
