﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Data.Services.Client;
using System.Linq;

namespace FlixPoll
{
    public class MovieViewModel : INotifyPropertyChanged
    {
        private List<LocalData.MovieTable> _movies;
        private List<LocalData.Quiz> _flixPollQuiz;

        public EventHandler<SyncDoneEventArgs> SyncDone;

        public List<LocalData.MovieTable> Movies
        {
            get
            {
                if (this._movies == null)
                    this._movies = new List<LocalData.MovieTable>();

                return this._movies;
            }
            set
            {
                if (this._movies != value)
                {
                    this._movies = value;
                    RaisePropertyChanged("Movies");
                }
            }
        }

        public List<LocalData.Quiz> FlixPollQuiz
        {
            get
            {
                if (this._flixPollQuiz == null)
                    this._flixPollQuiz = new List<LocalData.Quiz>();

                return this._flixPollQuiz;
            }
            set
            {
                if (this._flixPollQuiz != value)
                {
                    this._flixPollQuiz = value;
                    RaisePropertyChanged("FlixPollQuiz");
                }
            }
        }

        private void RaisePropertyChanged(string propertyName)
        {
            PropertyChangedEventHandler temp = this.PropertyChanged;

            if (temp != null)
                temp(this, new PropertyChangedEventArgs(propertyName));
        }

        public event PropertyChangedEventHandler PropertyChanged;

        public bool IsDataLoaded
        {
            get;
            private set;
        }

        // Get data and display on the page
        public void LoadData()
        {
            if (MoviesInLocalDb())
            {
                GetMoviesFromLocalDb();
            }
            else
            {
                GetMoviesFromService();
            }

            if (InquiriesInLocalDb())
            {
                GetInquiriesFromLocalDb();
            }
            else
            {
                GetInquiriesFromService();
            }
            this.IsDataLoaded = true;
        }

        private bool MoviesInLocalDb()
        {
            // see if there are any movies in the local CE db
            using (var flixPollDc =
                    new LocalData.FlixPollDataContext(App.DBConnectionString))
            {
                var movies =
                    from m in flixPollDc.Movies
                    select m.MovieId;

                return (movies.Count() > 0);
            }
        }

        private bool InquiriesInLocalDb()
        {
            // see if there are any questions in the local CE db
            using (var flixPollDc =
                    new LocalData.FlixPollDataContext(App.DBConnectionString))
            {
                var inquiries = from i in flixPollDc.Inquiries
                                select i.InquiryId;

                return (inquiries.Count() > 0);
            }
        }

        private void GetMoviesFromLocalDb()
        {
            using (var flixPollDc =
                    new LocalData.FlixPollDataContext(App.DBConnectionString))
            {
                // populate the movie list
                var movies =
                    from m in flixPollDc.Movies
                    select m;

                this.Movies = movies.ToList();
            }
        }

        private void GetInquiriesFromLocalDb()
        {
            // create a new quiz. When saved later, we'll add the movie id.
            // this model is bound to the form.
            using (var flixPollDB =
                    new LocalData.FlixPollDataContext(App.DBConnectionString))
            {
                var q =
                  from i in flixPollDB.Inquiries
                  select new LocalData.Quiz
                  {
                      Answer = false,
                      InquiryId = i.InquiryId,
                      Question = i.Question
                  };

                this.FlixPollQuiz = q.ToList();
            }
        }

        public void GetMoviesFromService()
        {
            // create a client context for the service
            var ctx = new FlixPollODataService.FlixPollOnCloudEntities(
                       new Uri(App.FlixPollServiceUrl));

            // define the query (deferred execution)
            var q =
                from m in ctx.Movies
                select m;

            // load the movies asynchronously
            var movies = new DataServiceCollection<FlixPollODataService.Movie>();
            movies.LoadCompleted += (o, e) =>
            {
                // Executes when the query completes
                if (e.Error != null)
                    throw e.Error;

                // once the data is retrieved, save it to local db
                CacheInLocalDb(movies);

                // once we have it locally, populate the list 
                GetMoviesFromLocalDb();
            };

            // begin the async query
            movies.LoadAsync(q);
        }

        private void GetInquiriesFromService()
        {
            // create a client context for the service
            var ctx = new FlixPollODataService.FlixPollOnCloudEntities(
                       new Uri(App.FlixPollServiceUrl));

            // deferred execution
            var q =
                from i in ctx.Inquiries
                select i;

            var inquiry = new DataServiceCollection<FlixPollODataService.Inquiry>(ctx);
            inquiry.LoadCompleted += (o, e) =>
            {
                if (e.Error != null)
                    throw e.Error;

                // once the data is retrieved, save it to local db
                CacheInLocalDb(inquiry);

                // once we have it locally, populate the list 
                GetInquiriesFromLocalDb();
            };

            // begin the async query
            inquiry.LoadAsync(q);
        }

        private void CacheInLocalDb(IEnumerable<FlixPollODataService.Movie> movies)
        {
            using (var flixPollDc =
                    new LocalData.FlixPollDataContext(App.DBConnectionString))
            {
                // first remove old movies from local db
                flixPollDc.Movies.DeleteAllOnSubmit(flixPollDc.Movies);

                // next, cache the movies in local db
                foreach (FlixPollODataService.Movie m in movies)
                {
                    LocalData.MovieTable newMovie = new LocalData.MovieTable();
                    newMovie.MovieId = m.MovieId;
                    newMovie.Name = m.Name;

                    flixPollDc.Movies.InsertOnSubmit(newMovie);
                }
                flixPollDc.SubmitChanges();
            }
        }

        private void CacheInLocalDb(IEnumerable<FlixPollODataService.Inquiry> inquiries)
        {
            using (var flixPollDc =
                    new LocalData.FlixPollDataContext(App.DBConnectionString))
            {
                // first remove old questions from local db
                flixPollDc.Inquiries.DeleteAllOnSubmit(flixPollDc.Inquiries);

                // next, cache the questions in local db
                foreach (FlixPollODataService.Inquiry q in inquiries)
                {
                    LocalData.InquiryTable newQuestion = new LocalData.InquiryTable();
                    newQuestion.InquiryId = q.InquiryId;
                    newQuestion.Question = q.Question;

                    flixPollDc.Inquiries.InsertOnSubmit(newQuestion);
                }
                flixPollDc.SubmitChanges();
            }
        }

        public void NewQuiz(int movieId)
        {
            using (var flixPollDc =
                    new LocalData.FlixPollDataContext(App.DBConnectionString))
            {
                // determine the next new response id. start from -1 and work backwards
                int newId = (flixPollDc.Responses.Count() > 0 ?
                              flixPollDc.Responses.Min(r => r.ResponseId) :
                              0);

                // insert each response into the SQL CE db
                foreach (LocalData.Quiz quiz in this.FlixPollQuiz)
                {
                    newId--;
                    LocalData.ResponseTable response = new LocalData.ResponseTable();
                    response.Answer = quiz.Answer;
                    response.InquiryId = quiz.InquiryId;
                    response.MovieId = movieId;
                    response.ResponseId = newId;

                    flixPollDc.Responses.InsertOnSubmit(response);

                    // reset the questions to be asked again
                    quiz.Answer = false;
                }
                flixPollDc.SubmitChanges();
            }
        }

        public void SaveMovieToLocalDb()
        {
            using (var flixPollDc =
                    new LocalData.FlixPollDataContext(App.DBConnectionString))
            {
                // save any movie name changes
                List<LocalData.MovieTable> changes =
                 this.Movies.Where(m => m.MovieChanged == true).ToList();

                if (changes != null && changes.Count > 0)
                {
                    foreach (LocalData.MovieTable c in changes)
                    {
                        var localMovieRecord =
                           (from m in flixPollDc.Movies
                            where m.MovieId == c.MovieId
                            select m).Single();

                        localMovieRecord.Name = c.Name;
                        localMovieRecord.UpdateToCloud = true;  // flag the record as changed.
                        // (will be updated to SQL Azure during sync)

                        // reset the ischanged flag so record is not constantly updated
                        c.MovieChanged = false;
                    }
                }
                // save the changes to local db.
                // changes will be uploaded to SQL Azure during sync process
                flixPollDc.SubmitChanges();
            }
        }

        private void RaiseSyncDone(SyncDoneEventArgs e)
        {
            EventHandler<SyncDoneEventArgs> temp = this.SyncDone;

            if (temp != null)
            {
                temp(this, e);
            }
        }

        public void UpdateToCloud()
        {
            using (var flixPollDc =
                    new LocalData.FlixPollDataContext(App.DBConnectionString))
            {
                // Any changed movies stored locally need to be updated to SQL Azure
                var q =
                    from m in flixPollDc.Movies
                    where m.UpdateToCloud == true
                    select m;

                foreach (LocalData.MovieTable localMovieRecord in q.ToArray())
                {
                    UpdateMovieInCloud(localMovieRecord);
                }
            }
            SendResponsesToCloud();
        }

        private void UpdateMovieInCloud(LocalData.MovieTable localMovieRecord)
        {
            // create a client context for the service
            var ctx = new FlixPollODataService.FlixPollOnCloudEntities(
                       new Uri(App.FlixPollServiceUrl));

            // get the movie from the service
            var q = from m in ctx.Movies
                    where m.MovieId == localMovieRecord.MovieId
                    select m;

            var dsq = q as DataServiceQuery<FlixPollODataService.Movie>;

            dsq.BeginExecute(r =>
            {
                var result = r.AsyncState as DataServiceQuery<FlixPollODataService.Movie>;
                var cloudMovieRecord = result.EndExecute(r).Single();

                if (cloudMovieRecord != null)
                {
                    if (localMovieRecord.MovieId == cloudMovieRecord.MovieId)
                    {
                        // Update the movie w/ the new name
                        cloudMovieRecord.Name = localMovieRecord.Name;

                        ctx.UpdateObject(cloudMovieRecord);

                        // push the changes back up to the service
                        ctx.BeginSaveChanges(CloudSync_Completed, ctx);
                    }
                }
            }, dsq);
        }

        private void SendResponsesToCloud()
        {
            using (var flixPollDc =
                    new LocalData.FlixPollDataContext(App.DBConnectionString))
            {
                // We are only storing NEW responses; just insert into the db
                var q =
                    from r in flixPollDc.Responses
                    select r;

                InsertToCloud(q.ToList());
            }
        }

        private void InsertToCloud(List<LocalData.ResponseTable> responses)
        {
            // create a client context for the service
            var ctx = new FlixPollODataService.FlixPollOnCloudEntities(
                       new Uri(App.FlixPollServiceUrl));
            
            bool hasChanges = (responses.Count() > 0);

            foreach (LocalData.ResponseTable localResponse in responses)
            {
               // hasChanges = true;
                FlixPollODataService.Response newResponse =
                 new FlixPollODataService.Response()
                 {
                     Answer = localResponse.Answer,
                     InquiryId = localResponse.InquiryId,
                     MovieId = localResponse.MovieId
                 };
                ctx.AddToResponses(newResponse);
            }

            if (hasChanges == true)
            {
                // async call to insert the new responses. perform the insert as a batch
                ctx.BeginSaveChanges(SaveChangesOptions.Batch, CloudSync_Completed, ctx);
            }
        }

        private void CloudSync_Completed(IAsyncResult result)
        {
            var ctx =
              result.AsyncState as FlixPollODataService.FlixPollOnCloudEntities;

            DataServiceResponse response = ctx.EndSaveChanges(result);

            // TODO: a better idea is to use a progress bar and check result for errors
            this.RaiseSyncDone(new SyncDoneEventArgs { Success = true });
        }

        public void ResetMoviesAfterSync()
        {
            using (var flixPollDc =
                    new LocalData.FlixPollDataContext(App.DBConnectionString))
            {
                // reset UpdateToCloud flag in local db after updating cloud
                foreach (LocalData.MovieTable m in flixPollDc.Movies)
                {
                    m.UpdateToCloud = false;
                }
                flixPollDc.SubmitChanges();
            }
        }

        public void DeleteResponsesAfterSync()
        {
            using (var flixPollDc =
                    new LocalData.FlixPollDataContext(App.DBConnectionString))
            {
                // remove responses from local db after updating cloud
                flixPollDc.Responses.DeleteAllOnSubmit(flixPollDc.Responses);
                flixPollDc.SubmitChanges();
            }
        }
    }
}
